/*
	Sherlock reporting routines.

	source:  sl_dump.c
	started: November 4, 1993.
	version: February 5, 1996.
		November 7, 1995.
			Call l2s instead of sprintf.
		September 26, 1995.
			Bug fix to sl_dump_start.
			Several minor formatting changes.
		December 26, 1993.
			Installed variable width columns in dumps.
*/
#include <LIBlib.h>

#include <sl.h>
#include <sl2.h>

#include <stdlib.h>
#include <string.h>

#undef USE_TOTAL_TIME

/*
	Define the default widths of the various columns.
	These widths are actually recomputed in sl_dump_start.
*/
long TRACEPOINT_WIDTH = 18;
long CALL_WIDTH = 15;
long OWN_WIDTH = 17;
long INNER_WIDTH = 17;
long ENABLED_WIDTH = 10;

/*
	The fudge factor is an additional "by hand" adjustment.
	It is used to compensate for the fact that the Sherlock timing macros
	cut off the ticks as soon possible, but that understates the Sherlock
	overhead, especially for often-called routines.
*/

/* Define global variables. */
bool sl_use_adjust_flag = TRUE;

/* Define local variables. */

static double sl_adjval;		/* Time fudge factor. */

static long tot_times;			/* Raw statistics. */
static long tot_2times;
static long tot_calls;
static long tot_ttimes;

static long totp_calls;			/* Percentages corresponding to raw statistics. */
static long totp_times;
static long totp_2times;
static long totp_ttimes;

static long subtot_calls;		/* Subtotals. */
static long subtotp_calls;
static long subtot_times;
static long subtotp_times;
static long subtot_2times;
static long subtotp_2times;

static long runtot_calls;		/* Running totals used to print grand totals. */
static long runtot_times;
static long runtot_2times;

static	sl_node a_list;			/* Head of alphabetical list of nodes. */

/*
	Function prototypes for internal routines.
*/
static unsigned long
	sl_adjust (sl_node * np, long val);
static void
	sl_dump_col (int width, long n1, long n2, long n3,char * delim, int msec_flag);
static short
	l2s (char * buf, long val);

/*
	This routine defines the meaning of the sl_adjval constant.

	return val - (long) (p->calls * sl_adjval).
*/
static unsigned long
sl_adjust(sl_node * np, long val)
{
	double dval;
	register long rval;
	static bool adj_flag = FALSE;

	if (sl_use_adjust_flag) {

		/* 2/18/93: Set sl_adjval on the fly. */
		if (adj_flag == FALSE) {
			sl_setadj();
			adj_flag = TRUE;
		}

		dval = (double) (np -> calls);
		dval = dval * sl_adjval;
		rval  = val - (long) dval;
		if (rval < 0) {
			rval = 0;
		}
		return rval;
	}
	else {
		return val;
	}
}

/*
	SL_DUMP macro--Print all statistics.
*/

void
sl_dump(void)
{
	sl_dump_start();
	sl_dump_header(SL_DUMP_ALL_STATS);
	sl_dump_names(SL_DUMP_ALL_STATS, NULL);
	sl_dump_totals();
	sl_dump_wild();
}

/*
	Dump only enabled entries.
*/

void
sl_edump(void)
{
	sl_dump_start();
	sl_dump_header(SL_DUMP_ENABLED_STATS);
	sl_dump_names(SL_DUMP_ENABLED_STATS, NULL);
	sl_dump_totals();
}

/*
	Print one column of statistics.
*/
static void
sl_dump_col(
	int width, long n1, long n2, long n3,
	char * delim, int msec_flag)
{
	char buf [CVT_BUF_SIZE];

	if (n1 < 0) n1 = 0;
	if (n2 < 0) n2 = 0;
	if (n3 < 0) n3 = 0;

	if (msec_flag) {
		epadmsec(n1, width);
	}
	else {
		epadlong(n1, width);
	}
	es(" ");
	es(delim);
	epads(cvt_percent(buf, CVT_BUF_SIZE, n2, n3) , 5);
	es("%");
}

/*
	Print the header used by sl_dump and sl_edump.
*/
void
sl_dump_header(bool dump_all_flag)
{
	ecnls(2);
	es(dump_all_flag ? "All " : "Enabled ");
	es("Sherlock statistics.  (Times in msec.)\n");
	ecnls(2);

	epads("tracepoints",	TRACEPOINT_WIDTH);
	epads("calls",			CALL_WIDTH+8);
	epads("own time",		OWN_WIDTH+8);
	epads("inner time",		INNER_WIDTH+8);
	epads("enabled", ENABLED_WIDTH);
	enl(); enl();
}

/*
	Dump the statistics for one name.

	Do *not* initialize the statistic here.
*/
void
sl_dump_name(char * name)
{
	sl_dump_names(FALSE, name);
}

/*
	Dump all the names with interesting statistics.
	If name is specified, dump only that name.
*/
void
sl_dump_names(bool dump_all_flag, char * dump_name)
{
	register sl_node	* p;

	for (p = a_list.alpha; p; p = p -> alpha) {
		if (
			(dump_name != NULL && strcmp(p -> name, dump_name)==0) ||
			(
				dump_name == NULL && dump_all_flag &&
				(p -> calls || p -> time || p -> time2)
			)
		) {

			/* Print breakpoint name. */	
			epads(p -> name, TRACEPOINT_WIDTH);

			sl_dump_col(CALL_WIDTH, p -> calls, p -> calls,
				tot_calls, "=", FALSE);

			runtot_calls += p -> calls;
			subtot_calls += p -> calls;

			totp_calls    += cvt_1000percent(p -> calls, tot_calls);
			subtotp_calls += cvt_1000percent(p -> calls, tot_calls);
			{
				long n = sl_adjust(p, p -> time);
				sl_dump_col(OWN_WIDTH, n, n, tot_times, "=", TRUE);
			}
			{
				long n = sl_adjust(p, p -> time);
				runtot_times += n;
				subtot_times += n;
			}
			{
				long n = cvt_1000percent(sl_adjust(p, p -> time), tot_times);
				totp_times += n;
				subtotp_times += n;
			}
			{
				long n = sl_adjust(p, p -> time2);
				sl_dump_col(INNER_WIDTH, n, n, tot_2times, "=", TRUE);
			}
			{
				long n = sl_adjust(p, p -> time2);
				runtot_2times += n;
				subtot_2times += n;
			}
			{
				long n = cvt_1000percent(sl_adjust(p, p -> time2), tot_2times);
				totp_2times += n;
				subtotp_2times += n;
			}

			/* Indicate whether tracing was enabled. */
			if (dump_all_flag || dump_name != NULL) {
				epadbool(p -> trace, ENABLED_WIDTH);
			}
			enl();
		}
	}
}

void
sl_dump_start(void)
{
	register sl_node	* p;
	register char	* pname;
	register sl_node	* q;
	register sl_node * q1;
	register int		i;
	char buf [CVT_BUF_SIZE];

	/* Be very careful. */
	if (sl_htab == NULL) {
		return;
	}

	/*
		Initialize alpabetical list.
	*/
	a_list.alpha = NULL;
	a_list.next  = NULL;
	a_list.name  = NULL;

	/* Calculate grand total times. */
	tot_calls = 0;
	tot_times = 0;
	tot_2times = 0;
	tot_ttimes = 0;

	totp_calls = 0;
	totp_times = 0;
	totp_2times = 0;
	totp_ttimes = 0;

	runtot_calls = 0;
	runtot_times = 0;
	runtot_2times = 0;

	/*
		Bug fix: 9/26/95: make sure the columns are wide enough for their headings.
		This code is ugly because it depends on sl_dump_col.
	*/
	ENABLED_WIDTH		= 2 + strlen("enabled");
	TRACEPOINT_WIDTH	= 2 + strlen("tracepoints");
	CALL_WIDTH			= 4;
	OWN_WIDTH			= 4;
	INNER_WIDTH			= 4;

	/* Scan all buckets. */
	for (i = 0; i < SL_MAX_HASH; i++) {

		for (p = sl_htab[i]; p; p = p -> next) {

			tot_calls  += p -> calls;
			tot_times  += sl_adjust(p, p -> time);
			tot_2times += sl_adjust(p, p -> time2);

			/*
				Sort the entries that will be printed by creating
				a list in the alpha field of the nodes.
			*/
			if (p -> calls || p -> time || p -> time2) {
				pname = p -> name;
				TRACEPOINT_WIDTH = max(TRACEPOINT_WIDTH, strlen(p -> name));
				for (q1 = &a_list, q = q1 -> alpha;
					;
					 q1 = q, q = q -> alpha
					) {

					if (q == NULL ||
						strcmp(pname, q -> name) < 0
					   ) {
						q1 -> alpha = p;
						p  -> alpha = q;
						break;
					}
				}
			}
		}
	}

	tot_ttimes = tot_times + tot_2times;
	
	#if 0 /* Don't call sprintf */

		CALL_WIDTH	= max(CALL_WIDTH,	1 + sprintf(buf, "%ld", tot_calls));
		OWN_WIDTH	= max(OWN_WIDTH,	1 + sprintf(buf, "%ld", ticks2msec(tot_times)));
		INNER_WIDTH	= max(INNER_WIDTH,	1 + sprintf(buf, "%ld", ticks2msec(tot_2times)));
	
	#else
	
		CALL_WIDTH	= max(CALL_WIDTH,	1 + l2s(buf, tot_calls));
		OWN_WIDTH	= max(OWN_WIDTH,	1 + l2s(buf, ticks2msec(tot_times)));
		INNER_WIDTH	= max(INNER_WIDTH,	1 + l2s(buf, ticks2msec(tot_2times)));
	
	#endif
}

/*
	Print wildcard table.
*/
void
sl_dump_wild(void)
{
	register sl_node	* p;

	ecnls(2);
	epads("Wildcards", 20); es("   tracing");

	for (p = sl_wcard; p; p = p -> next) {
		enl();
		epads(p -> name, 20);
		eblanks(6);
		ebool(p -> trace);
	}
	enl();
}

/*
	Print the subtotals.
*/
void
sl_dump_subtotals(void)
{
	ecnl();
	epads("SUB TOTALS", TRACEPOINT_WIDTH);
	sl_dump_col(CALL_WIDTH,  subtot_calls,  subtotp_calls,  1000, " ", FALSE);
	sl_dump_col(OWN_WIDTH,   subtot_times,  subtotp_times,  1000, " ", TRUE);
	sl_dump_col(INNER_WIDTH, subtot_2times, subtotp_2times, 1000, " ", TRUE);
	ecnls(2);
}

void
sl_dump_subtotal_start(void)
{
	subtot_calls = 0;
	subtotp_calls = 0;
	subtot_times = 0;
	subtotp_times = 0;
	subtot_2times = 0;
	subtotp_2times = 0;
}

/*
	Print the totals.
*/
void
sl_dump_totals(void)
{
	ecnl();
	epads("TOTALS:", TRACEPOINT_WIDTH);
	sl_dump_col(CALL_WIDTH,  runtot_calls,  totp_calls,  1000, " ", FALSE);
	sl_dump_col(OWN_WIDTH,   runtot_times,  totp_times,  1000, " ", TRUE);
	sl_dump_col(INNER_WIDTH, runtot_2times, totp_2times, 1000, " ", TRUE);

	ecnls(2);
	es("Sherlock nodes: "); eint(sl_node_count);
	es(" SL_MAX_NODES: "); eint(SL_MAX_NODES); enl();
	if (sl_neg_count > 0) {
		es("Negative times: "); elong(sl_neg_count);
	}

#if 0 // defined(THINK_C) // Doesn't work on PowerPc.
	ecnl();
	es("maximum stack size: "); elong(sl_margin_start - sl_min_margin);
	es(" minimum stack margin: "); elong(sl_min_margin); enl();
#endif

	#if 1 /* This forces the whole sprintf logic to be linked in. */
		if (sl_use_adjust_flag) {
			es("Times adjusted by "); edouble(sl_adjval); es(" (ticks/call)"); enl();
		}
	#endif

	ecnl();
}

/*
	Print a report of the usage of the bins.
*/
void
sl_hdump(void)
{
	register int		i;
	register int		bucket_count;
	register sl_node	*p;
	long bin [10];

	/* Be very careful. */
	if (sl_htab == NULL) {
		return;
	}

	for (i = 0; i < 10; i++) {
		bin[i] = 0;
	}

	/* Scan all buckets. */
	for (i = 0; i < SL_MAX_HASH; i++) {

		bucket_count = 0;

		for (p = sl_htab[i]; p; p = p -> next) {
			bucket_count++;
		}

		if (bucket_count >= 9) {
			bin[9]++;
		}
		else {
			bin[bucket_count]++;
		}
	}

	es("\nHash statistics\n\n");

	for (i = 0; i < 10; i++) {
		es("bin["); eint(i);
		es("] = "); elong(bin[i]); enl();
	}
	enl();
}

/*
	Dump the call stack.
	Very useful for debugging.
*/
void
sl_dumpstk(void)
{
	register int 		i;
	register sl_snode	* sp;
	register sl_node	* np;

	/* Make sure there *is* a call stack. */
	if (sl_stack == NULL) {
		return;
	}

	ecnls(2); es("Dump of call stack:");

	for (i = 0; i < sl_level; i++) {

		ecnl();
		sp = &sl_stack[i];
		np = sp -> node;

		es("name: ");	 epads(np -> name, 20);
		es(" stat: ");  epadlong(np -> calls, 4);
		es(" time: ");  epadlong(np -> time,  4);
		es(" time2: "); epadlong(np -> time2, 4);
		enl();
	}

	ecnls(2); es("Dump of recent call stack:");

	for (i = sl_level; i < sl_lmax; i++) {

		sp = &sl_stack[i];
		np = sp -> node;

		ecnl();
		es("name: ");	epads(np -> name, 20);
		es(" calls: "); epadlong(np -> calls, 4);
		es(" time: ");  epadlong(np -> time,  4);
		es(" time2: "); epadlong(np -> time2, 4);
		enl();
	}
	ecnl();
}

/*
	Set sl_adjval to the "asymptotic" Sherlock overhead.

	This overhead does *not* include the time spent searching for the
	tracepoint the first time the macro is encountered.

	The value of sl_adjval does not affect how statistics are gathered,
	only how the routines in this file report the statistics already gathered.
*/

static void dummy(int i);

/* Make sure the STAT macro are always defined. */
#define STAT2(a)	{static sl_node *sl_h_=0; sl2stat(&sl_h_,a);}
#define STATB2(a)	{static sl_node *sl_h_=0; sl2bstat(&sl_h_,a);}
#define STATX2(a)	{sl_xtrace(a);}

void
sl_setadj(void)
{
	unsigned long start;
	unsigned long val1;
	unsigned long val2;
	register int i;
	double d1000;
	double dval;

	/* Get value with no Sherlock overhead. */

	start = sl_ticks();
	for (i = 0; i < 1000; i++) {
		dummy(i);
	}
	val1 = sl_ticks() - start;

	/*
		We will perform this loop 1000 times so that we can
		get a better value for the average value of the overhead.
		(The constant 1000 appears in the calculation below.)

		A key point:  the sl_adjval is used to subtract a "fudge"
		factor from timing statistics based on the "calls" statistic.
		But only the STATB macro increments the "calls" statistics,
		so we have to have *both* a STATB and a STATX macro in the following
		loop to get a proper estimate of the Sherlock overhead.
	*/
	
	#ifdef SL_USE_MICROSECONDS
	{
		UnsignedWide wide_time;
		Microseconds(&wide_time);
		start = wide_time.lo;
	}
	#else
		start = sl_ticks();
	#endif
	
	for (i = 0; i < 1000; i++) {
		STATB2("sl_setadj");
		dummy(i);
		STATX2("sl_setadj");
	}

	
	#ifdef SL_USE_MICROSECONDS
	{
		UnsignedWide wide_time;
		Microseconds(&wide_time);
		val2 = wide_time.lo - start;
	}
	#else
			val2 = sl_ticks() - start;
	#endif

	dval = val2 - val1;
	d1000 = (double) 1000;
	sl_adjval = (dval / d1000);
	if (sl_adjval < 0.0) {
		sl_adjval = 1.0;
	}
	
	// Clear the statistics for sl_set_adj
	{
		sl_node * sp = sl_find("sl_setadj", "sl_setadj");
		sp -> calls = 0;
		sp -> time = 0;
		sp -> time2 = 0;
	}
}

/*
	A do-nothing routine that may prevent some unwanted loop optimizations.
	The speed of this routine should not affect the value of TIME_ADJUST.
*/
static int global_var;

static void
dummy(int i)
{
	global_var += i;
}

/*
	l2s(buf, val); is the same as sprintf(buf, "%ld", val);
	We define l2s to avoid pulling in the whole floating point library.
*/
static short
l2s (char * buf, long val)
{
	short length = 0;
	
	if (val < 0) {
		buf[length++] = '-';
		val = -val;				/* A small bug: this can cause overflow. */
	}
	
	/* Generate the digits in reverse order. */
	do {
		buf[length++] = (val % 10) + '0';
		val /= 10;
	}
	while (val > 0);
	
	/* Reverse the digits. */
	{
		short i = (buf[0] == '-');
		short j = length-1;
		
		while (i < j) {
			char temp = buf[j];
			buf[j] = buf[i];
			buf[i] = temp;
			i++; j--;
		}
	}
	
	buf[length] = '\0';
	return length;
}